// Timothy Hanson
// Knight's Tour
/******************************************************************
User inputs a value n which creates a chessboard of n*n dimension
and a starting position (x,y) for the knight.  The problem is to
compute n^(2) - 1 moves (if such a tour exists) in which the knight
visits every square on the board once and ONLY once.
*******************************************************************/

#include <iostream>    // For user input
#include <iomanip>     // For some clean formatting
#include "d_matrix.h"  // Used to setup the board
#include "d_except.h"  // d_matrix.h needs it

// Function Prototypes
bool knight(matrix<int>& board, int k, int i, int j, int& calls);
void drawBoard(matrix<int>& board);

using namespace std;

int main ()
{
	int n=0,         // n^2 will create the board
		calls=0,	 // Let's see how many recursive calls we made
		xpos=0,      // Starting X-Position
		ypos=0,      // Starting Y-Position
		k=1;		 // Move number, k will eventually = number of moves required to solve problem
					 // which also happens to be the number of squares on the board, or n^2
	char ch = 'y';	 // Use this to loop the program until the user gets bored

	// User Welcome/Warning
	cout << "Knight's Tour Algorithm:" << endl;
	cout << "For large board sizes, search times for solutions will be significantly longer."
		<< endl << "The coordinates of the board start at 0 and go to the specified input." << endl;

	while(ch=='y' || ch=='Y')
	{

		// User Input
		cout << endl << "Enter Board Size (Number will be squared): ";
		cin >> n;
		cout << "Enter Starting x Position (Number must be less than board size): ";
		cin >> xpos;
		cout << "Enter Starting y Position (Number must be less than board size): ";
		cin >> ypos;

		// Reinitialize for when main loop is run again
		calls = 0;

		// Check for user error
		if(xpos>n-1 || ypos>n-1 || xpos<0 || ypos<0)
		{
			cout << "Error: Invalid input." << endl << endl;
			return 1;
		}

		// Spacer
		cout << endl;
	
		// Check input
		cout << "Number of squares on this board: " << n*n << endl;
		cout << "Starting position was: " << xpos << "," << ypos << endl << endl;

		// Create the matrix and fill with zeroes
		matrix<int> board(n,n,0);

		// Let them know what's going on
		cout << "Analyzing combinations..";

		// see if there is a solution
		if (knight(board, k, xpos, ypos, calls))
		{
			// Display the solution
			cout << endl << endl << "Tour Exists As Shown" << endl;
			drawBoard(board);
			cout << "Number of function calls made: " << calls << endl;
		}
		else
		{
			// No answer
			cout << endl << endl << "No solution with specified parameters." << endl;
			cout << "Number of function calls: " << calls << endl << endl;
		}
		
		// Will we dance again?
		cout << "Input another board? y/n: ";
		cin >> ch;
	}
	return 0;
}

// Returns a true/false answer if it's possible to
// make a tour with the given inputs.
// Variables taken in:
// board: The chessboard whose values get updated from value zero (not visited) to value k
// k: The kth move, or the move number that we're looking at
// i: x position on the board
//.j: y position on the board
// calls: Number of times this function has been called
bool knight(matrix<int>& board, int k, int i, int j, int& calls)
{
	bool result;								// What we want to know
	int u, v,									// u and v will be the new i and j positions
		m;										// m is the movement counter
	const int n = board.rows();					// Checked against k to see if we're done

	static int dx[] = {2,1,-1,-2,-2,-1, 1, 2},  // Used for checking possible knight movements
	           dy[] = {1,2, 2, 1,-1,-2,-2,-1};  // seperated into x and y components

	// Assume this square will be move k
	board[i][j] = k;

	// How many times have we hit this function
	calls++;

	// Let people know we haven't locked up
	if(calls%75000==0)
		cout << ".";

	if (k == n*n)
	{
		// That was the last move
		result = true;
	}
	else
	{
		// We got work to do
		result = false;
		m = 0;

		// look at all 8 possible moves and see if one leads to a solution
		while((!result) && (m<8))
		{
			//use dx, dy to determine (u,v)
			u = i + dx[m];
			v = j + dy[m];

			//If board[u][v] is on the board and has not been visited
			//try to find a solution by moving there
			if( (u>=0) && (u<n) && (v>=0) && (v<n) && board[u][v]==0)
				result=knight(board,k+1,u,v,calls);
			
			// Advance through the possible moves
			m++;
		}
		
		//If this position(i,j) did not work, assign 
		//0 to this board position and return false
		//otherwise we're done, return the result.
		if(!result)
			board[i][j] = 0;
	}
	return result;
}

// Draws the matrix in n*n dimensions
// Takes in the board matrix for output
// Currently no axis labels.  
void drawBoard(matrix<int>& board)
{
	int g, h, // For the dual loops
		q;    // For X Axis coordinates
	const int n = board.rows();  // So we write just enough rows/columns.

	// Spacer
	cout << endl << "     ";

	// Show X Axis Coordinates
	for(q=0; q<n; q++)
		cout << setw(4) << q;
	cout << endl << endl;


	// Loop through y column
	for (g=0;g<n;g++)
	{
		// Show Y Axis Coordinates
		cout<< setw(4) << g << " ";

		// Loop through x row
		for(h=0;h<n;h++)
			cout << setw(4) << board[h][g];
		cout << endl << endl;
	}

	// Spacer
	cout << endl;
}
/*********************************

Sample input:
Board: 5x5
Starting Position: (0,0)

        0   1   2   3   4
   
   0    1  14  19   8  25

   1    6   9   2  13  18

   2   15  20   7  24   3

   3   10   5  22  17  12

   4   21  16  11   4  23

*********************************/